@@ -39,6 +39,7 @@ const pfmlib_attr_desc_t intel_x86_mods[]={
 	PFM_ATTR_B("i", "invert"),				/* invert */
 	PFM_ATTR_I("c", "counter-mask in range [0-255]"),	/* counter-mask */
 	PFM_ATTR_B("t", "measure any thread"),			/* monitor on both threads */
+	PFM_ATTR_I("ldlat", "load latency threshold (cycles, [3-65535])"),	/* load latency threshold */
 	PFM_ATTR_NULL /* end-marker to avoid exporting number of entries */
 };
 
@@ -61,20 +62,6 @@ cpuid(unsigned int op, unsigned int *a, unsigned int *b, unsigned int *c, unsign
        : "a" (op));
 }
 
-static inline int
-is_model_umask(void *this, int pidx, int attr)
-{
-	pfmlib_pmu_t *pmu = this;
-	const intel_x86_entry_t *pe = this_pe(this);
-	const intel_x86_entry_t *ent;
-	unsigned int model;
-
-	ent = pe + pidx;
-	model = ent->umasks[attr].umodel;
-
-	return model == 0 || model == pmu->pmu;
-}
-
 static void
 pfm_intel_x86_display_reg(void *this, pfmlib_event_desc_t *e)
 {
@@ -125,46 +112,6 @@ intel_x86_num_mods(void *this, int idx)
 	return pfmlib_popcnt(mask);
 }
 
-static unsigned int
-intel_x86_num_umasks(void *this, int pidx)
-{
-	pfmlib_pmu_t *pmu = this;
-	const intel_x86_entry_t *pe = this_pe(this);
-	unsigned int i, n = 0, model;
-
-	/*
-	 * some umasks may be model specific
-	 */
-	for (i = 0; i < pe[pidx].numasks; i++) {
-		model = pe[pidx].umasks[i].umodel;
-		if (model && model != pmu->pmu)
-			continue;
-		n++;
-	}
-	return n;
-}
-
-/*
- * find actual index of umask based on attr_idx
- */
-int
-intel_x86_attr2umask(void *this, int pidx, int attr_idx)
-{
-	const intel_x86_entry_t *pe = this_pe(this);
-	unsigned int i;
-
-	for (i = 0; i < pe[pidx].numasks; i++) {
-
-		if (!is_model_umask(this, pidx, i))
-			continue;
-
-		if (attr_idx == 0)
-			break;
-		attr_idx--;
-	}
-	return i;
-}
-
 int
 intel_x86_attr2mod(void *this, int pidx, int attr_idx)
 {
@@ -380,13 +327,15 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 	const intel_x86_entry_t *pe;
 	pfm_intel_x86_reg_t reg;
 	unsigned int grpmsk, ugrpmsk = 0;
-	uint64_t umask1, umask2;
+	uint64_t umask1, umask2, ucode, last_ucode = ~0ULL;
 	unsigned int modhw = 0;
 	unsigned int plmmsk = 0;
+	int umodmsk = 0, modmsk_r = 0;
 	int k, ret, id;
 	unsigned int max_grpid = INTEL_X86_MAX_GRPID;
 	unsigned int last_grpid =  INTEL_X86_MAX_GRPID;
 	unsigned int grpid;
+	int ldlat = 0, ldlat_um = 0;
 	int grpcounts[INTEL_X86_NUM_GRP];
 	int ncombo[INTEL_X86_NUM_GRP];
 
@@ -409,6 +358,8 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 	umask1 = (reg.val >> 8) & 0xff;
 	umask2 = 0;
 
+	modmsk_r = pe[e->event].modmsk_req;
+
 	for (k = 0; k < e->nattrs; k++) {
 		a = attr(e, k);
 
@@ -451,6 +402,8 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 			if (intel_x86_uflag(this, e->event, a->idx, INTEL_X86_NCOMBO))
 				ncombo[grpid] = 1;
 
+			if (intel_x86_uflag(this, e->event, a->idx, INTEL_X86_LDLAT))
+				ldlat_um = 1;
 			/*
 			 * if more than one umask in this group but one is marked
 			 * with ncombo, then fail. It is okay to combine umask within
@@ -462,10 +415,21 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 			}
 
 			last_grpid = grpid;
+			ucode     = pe[e->event].umasks[a->idx].ucode;
 			modhw    |= pe[e->event].umasks[a->idx].modhw;
-			umask2   |= pe[e->event].umasks[a->idx].ucode >> 8;
+			umask2   |= ucode >> 8;
 			ugrpmsk  |= 1 << pe[e->event].umasks[a->idx].grpid;
 
+			modmsk_r |= pe[e->event].umasks[a->idx].umodmsk_req;
+
+			if (intel_x86_uflag(this, e->event, a->idx, INTEL_X86_CODE_OVERRIDE)) {
+				if (last_ucode != ~0ULL && (ucode & 0xff) != last_ucode) {
+					DPRINT("cannot override event with two different codes for %s\n", pe[e->event].name);
+					return PFM_ERR_FEATCOMB;
+				}
+				last_ucode = ucode & 0xff;
+				reg.sel_event_select = last_ucode;
+			}
 		} else if (a->type == PFM_ATTR_RAW_UMASK) {
 
 			/* there can only be one RAW_UMASK per event */
@@ -485,11 +449,13 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 					if (modhw & _INTEL_X86_ATTR_I)
 						return PFM_ERR_ATTR_SET;
 					reg.sel_inv = !!ival;
+					umodmsk |= _INTEL_X86_ATTR_I;
 					break;
 				case INTEL_X86_ATTR_E: /* edge */
 					if (modhw & _INTEL_X86_ATTR_E)
 						return PFM_ERR_ATTR_SET;
 					reg.sel_edge = !!ival;
+					umodmsk |= _INTEL_X86_ATTR_E;
 					break;
 				case INTEL_X86_ATTR_C: /* counter-mask */
 					if (modhw & _INTEL_X86_ATTR_C)
@@ -497,24 +463,34 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 					if (ival > 255)
 						return PFM_ERR_ATTR_VAL;
 					reg.sel_cnt_mask = ival;
+					umodmsk |= _INTEL_X86_ATTR_C;
 					break;
 				case INTEL_X86_ATTR_U: /* USR */
 					if (modhw & _INTEL_X86_ATTR_U)
 						return PFM_ERR_ATTR_SET;
 					reg.sel_usr = !!ival;
 					plmmsk |= _INTEL_X86_ATTR_U;
+					umodmsk |= _INTEL_X86_ATTR_U;
 					break;
 				case INTEL_X86_ATTR_K: /* OS */
 					if (modhw & _INTEL_X86_ATTR_K)
 						return PFM_ERR_ATTR_SET;
 					reg.sel_os = !!ival;
 					plmmsk |= _INTEL_X86_ATTR_K;
+					umodmsk |= _INTEL_X86_ATTR_K;
 					break;
 				case INTEL_X86_ATTR_T: /* anythread (v3 and above) */
 					if (modhw & _INTEL_X86_ATTR_T)
 						return PFM_ERR_ATTR_SET;
 					reg.sel_anythr = !!ival;
+					umodmsk |= _INTEL_X86_ATTR_T;
 					break;
+				case INTEL_X86_ATTR_LDLAT: /* load latency */
+					if (ival < 3 || ival > 65535)
+						return PFM_ERR_ATTR_VAL;
+					ldlat = ival;
+					break;
+
 			}
 		}
 	}
@@ -555,6 +531,10 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 		}
 	}
 
+	if (modmsk_r && (umodmsk ^ modmsk_r)) {
+		DPRINT("required modifiers missing: 0x%x\n", modmsk_r);
+		return PFM_ERR_ATTR;
+	}
 	/*
 	 * reorder all the attributes such that the fstr appears always
 	 * the same regardless of how the attributes were submitted.
@@ -579,6 +559,24 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 		e->count = 1;
 	}
 
+	if (ldlat && !ldlat_um) {
+		DPRINT("passed ldlat= but not using ldlat umask\n");
+		return PFM_ERR_ATTR;
+	}
+
+	/*
+	 * force a default ldlat (will not appear in display_reg)
+	 */
+	if (ldlat_um && !ldlat) {
+		DPRINT("missing ldlat= for umask, forcing to default %d cycles\n", INTEL_X86_LDLAT_DEFAULT);
+		ldlat = INTEL_X86_LDLAT_DEFAULT;
+	}
+
+	if (ldlat && ldlat_um) {
+		e->codes[1] = ldlat;
+		e->count = 2;
+	}
+
 	/* take into account hardcoded modifiers, so use or on reg.val */
 	reg.val     |= (umask1 | umask2)  << 8;
 
@@ -626,6 +624,9 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 		case INTEL_X86_ATTR_T:
 			evt_strcat(e->fstr, ":%s=%lu", intel_x86_mods[id].name, reg.sel_anythr);
 			break;
+		case INTEL_X86_ATTR_LDLAT:
+			evt_strcat(e->fstr, ":%s=%d", intel_x86_mods[id].name, ldlat);
+			break;
 		}
 	}
 	return PFM_SUCCESS;
@@ -634,16 +635,8 @@ pfm_intel_x86_encode_gen(void *this, pfmlib_event_desc_t *e)
 int
 pfm_intel_x86_get_encoding(void *this, pfmlib_event_desc_t *e)
 {
-	const intel_x86_entry_t *pe = this_pe(this);
 	int ret;
 
-	/*
-	 * If event requires special encoding, then invoke
-	 * model specific encoding function
-	 */
-	if (intel_x86_eflag(this, e->event, INTEL_X86_ENCODER))
-		return pe[e->event].encoder(this, e);
-
 	ret = pfm_intel_x86_encode_gen(this, e);
 	if (ret != PFM_SUCCESS)
 		return ret;
@@ -745,9 +738,9 @@ pfm_intel_x86_validate_table(void *this, FILE *fp)
 		for (j=i+1; j < (int)pmu->pme_count; j++) {
 			if (pe[i].code == pe[j].code && !(pe[j].equiv || pe[i].equiv) && pe[j].cntmsk == pe[i].cntmsk) {
 				fprintf(fp, "pmu: %s events %s and %s have the same code 0x%x\n", pmu->name, pe[i].name, pe[j].name, pe[i].code);
-			error++;
+				error++;
+				}
 			}
-		}
 
 		for(j=0; j < INTEL_X86_NUM_GRP; j++)
 			ndfl[j] = 0;
@@ -863,7 +856,11 @@ pfm_intel_x86_get_event_attr_info(void *this, int pidx, int attr_idx, pfm_event_
 		info->name = pe[pidx].umasks[idx].uname;
 		info->desc = pe[pidx].umasks[idx].udesc;
 		info->equiv= pe[pidx].umasks[idx].uequiv;
-		info->code = pe[pidx].umasks[idx].ucode >> 8; /* show actual umask code */
+
+		info->code = pe[pidx].umasks[idx].ucode;
+		if (!intel_x86_uflag(this, pidx, idx, INTEL_X86_CODE_OVERRIDE))
+			info->code >>= 8;
+
 		info->type = PFM_ATTR_UMASK;
 		info->is_dfl = intel_x86_uflag(this, pidx, idx, INTEL_X86_DFL);
 		info->is_precise = intel_x86_uflag(this, pidx, idx, INTEL_X86_PEBS);
@@ -949,3 +946,18 @@ pfm_intel_x86_get_event_nattrs(void *this, int pidx)
 	nattrs += intel_x86_num_mods(this, pidx);
 	return nattrs;
 }
+
+int
+pfm_intel_x86_can_auto_encode(void *this, int pidx, int uidx)
+{
+	int numasks;
+
+	if (intel_x86_eflag(this, pidx, INTEL_X86_NO_AUTOENCODE))
+		return 0;
+
+	numasks = intel_x86_num_umasks(this, pidx);
+	if (uidx >= numasks)
+		return 0;
+
+	return !intel_x86_uflag(this, pidx, uidx, INTEL_X86_NO_AUTOENCODE);
+}
